/*      This file is for functions having to do with key bindings,
 *      descriptions, help commands and startup file.
 *
 *      written 11-feb-86 by Daniel Lawrence
 *
 * Modifications by Thierry Godefroy:
 *
 *              03/12/98        QDOS/SMS port.
 *              11/12/98        M-X binding made unchangeable when compiled with
 *                               MENUS support (this binding is used by the menu
 *                               system).
 *              24/12/98        ^G binding made unchangeable as well (also used
 *                               by the menu system to abort commands).
 *              25/12/98        Same thing for ^M and ^[ !
 *              20/01/99        Added the EMACS_HOME environment variable in
 *                               flook() for QDOS (this prevents searching for
 *                               emacs files into DATA_USE when EMACS_HOME is
 *                               set).
 *                              help() now also seach in the HOME (EMACS_HOME or
 *                               DATA_USE for QDOS) directory and the test on
 *                               the buffer name is done using pathname[1]
 *                               instead of "emacs.hlp".
 *              01/03/99        flook() now accepts "," as a separator in PATH
 *                               for QDOS/SMS.
 *              31/10/99        Added "uelocal_rc" as a possible EMACS_HOME file
 *                               in flook() for QDOS.
 *              26/09/00        Added (char *) casts before the first parameter
 *                               of bfind() and flook() functions called from
 *                               help() and startup() so to remove gcc warnings. 
 */

#include        <stdio.h>
#include        "estruct.h"
#include        "eproto.h"
#include        "edef.h"
#include        "elang.h"
#include        "epath.h"

#if QDOS
#include        <qdos.h>
#endif

/* give me some help!!!!
 * bring up a fake buffer and read the help file
 * into it with view mode
 */
#if PROTO
int PASCAL NEAR help(int f, int n)
#else
int PASCAL NEAR help(f, n)
int f,n;        /* prefix flag and argument */
#endif
{
        register BUFFER *bp;    /* buffer pointer to help */
        char *fname;            /* file name of help file */

        /* first check if we are already here */
        bp = bfind((char *)pathname[1], FALSE, BFINVS);

        if (bp == NULL) {
                fname = flook((char *)pathname[1], TRUE);
                if (fname == NULL) {
                        mlwrite(TEXT12);
/*                              "[Help file is not online]" */
                        return(FALSE);
                }
        }

        /* split the current window to make room for the help stuff */
        if (splitwind(FALSE, 1) == FALSE)
                        return(FALSE);

        if (bp == NULL) {
                /* and read the stuff in */
                if (getfile(fname, FALSE) == FALSE)
                        return(FALSE);
        } else
                swbuffer(bp);

        /* make this window in VIEW mode, update all mode lines */
        curwp->w_bufp->b_mode |= MDVIEW;
        curwp->w_bufp->b_flag |= BFINVS;
        upmode();
        return(TRUE);
}

/* describe the command for a certain key */
#if PROTO
int PASCAL NEAR deskey(int f, int n)
#else
int PASCAL NEAR deskey(f, n)
int f,n;        /* prefix flag and argument */
#endif
{
        register int c;         /* key to describe */
        register char *ptr;     /* string pointer to scan output strings */
        char outseq[NSTRING];   /* output buffer for command sequence */

        /* prompt the user to type us a key to describe */
        mlwrite(TEXT13);
/*              ": describe-key " */

        /* get the command sequence to describe
           change it to something we can print as well */
        /* and dump it out */
        ostring(cmdstr(c = getckey(FALSE), &outseq[0]));
        ostring(" ");

        /* find the right ->function */
        if ((ptr = getfname(getbind(c))) == NULL)
                ptr = "Not Bound";

        /* output the command sequence */
        ostring(ptr);
        return(TRUE);
}

/* bindtokey:   add a new key to the key binding table          */
#if PROTO
int PASCAL NEAR bindtokey(int f, int n)
#else
int PASCAL NEAR bindtokey(f, n)
int f, n;       /* command arguments [IGNORED] */
#endif
{
        register unsigned int c;/* command key to bind */
        register int (PASCAL NEAR *kfunc)();/* ptr to the requested function to bind to */
        register KEYTAB *ktp;   /* pointer into the command table */
        register int found;     /* matched command flag */
        char outseq[80];        /* output buffer for keystroke sequence */

        /* prompt the user to type in a key to bind */
        /* get the function name to bind it to */
        kfunc = getname(TEXT15);
/*                      ": bind-to-key " */
        if (kfunc == NULL) {
                mlwrite(TEXT16);
/*                      "[No such function]" */
                return(FALSE);
        }
        if (clexec == FALSE) {
                ostring(" ");
                TTflush();
        }

        /* get the command sequence to bind */
        c = getckey((kfunc == meta) || (kfunc == cex) ||
                    (kfunc == unarg) || (kfunc == ctrlg));

        if (clexec == FALSE) {

                /* change it to something we can print as well */
                /* and dump it out */
                ostring(cmdstr(c, &outseq[0]));
        }

#if QDOS & MENUS
        if (!strcmp(outseq, "M-X") || !strcmp(outseq, "^G") ||
            !strcmp(outseq, "^M")  || !strcmp(outseq, "^[")) {
                mlwrite(TEXT401);
/*                     "Error: This binding is reserved for the menu system." */
                return(FALSE);
        }
#endif

        /* if the function is a unique prefix key */
        if (kfunc == unarg || kfunc == ctrlg || kfunc == quote) {

                /* search for an existing binding for the prefix key */
                ktp = &keytab[0];
                while (ktp->k_type != BINDNUL) {
                        if (ktp->k_type == BINDFNC && ktp->k_ptr.fp == kfunc)
                                unbindchar(ktp->k_code);
                        ++ktp;
                }

                /* reset the appropriate global prefix variable */
                if (kfunc == unarg)
                        reptc = c;
                if (kfunc == ctrlg)
                        abortc = c;
                if (kfunc == quote)
                        quotec = c;
        }

        /* search the table to see if it exists */
        ktp = &keytab[0];
        found = FALSE;
        while (ktp->k_type != BINDNUL) {
                if (ktp->k_code == c) {
                        found = TRUE;
                        break;
                }
                ++ktp;
        }

        if (found) {    /* it exists, just change it then */
                ktp->k_ptr.fp = kfunc;
                ktp->k_type = BINDFNC;
        } else {        /* otherwise we need to add it to the end */
                /* if we run out of binding room, bitch */
                if (ktp >= &keytab[NBINDS]) {
                        mlwrite(TEXT17);
/*                              "Binding table FULL!" */
                        return(FALSE);
                }

                ktp->k_code = c;        /* add keycode */
                ktp->k_ptr.fp = kfunc;  /* and the function pointer */
                ktp->k_type = BINDFNC;  /* and the binding type */
                ++ktp;                  /* and make sure the next is null */
                ktp->k_code = 0;
                ktp->k_type = BINDNUL;
                ktp->k_ptr.fp = NULL;
        }

        /* if we have rebound the meta key, make the
           search terminators follow it                 */
        if (kfunc == meta) {
                sterm = c;
                isterm = c;
        }

        return(TRUE);
}

/* macrotokey:  Bind a key to a macro in the key binding table */
#if PROTO
int PASCAL NEAR macrotokey(int f, int n)
#else
int PASCAL NEAR macrotokey(f, n)
int f, n;       /* command arguments [IGNORED] */
#endif
{
        register unsigned int c;/* command key to bind */
        register BUFFER *kmacro;/* ptr to buffer of macro to bind to key */
        register KEYTAB *ktp;   /* pointer into the command table */
        register int found;     /* matched command flag */
        register int status;    /* error return */
        char outseq[80];        /* output buffer for keystroke sequence */
        char bufn[NBUFN];       /* buffer to hold macro name */

        /* get the buffer name to use */
        if ((status=mlreply(TEXT215, &bufn[1], NBUFN-2)) != TRUE)
/*              ": macro-to-key " */
                return(status);

        /* build the responce string for later */
        strcpy(outseq, TEXT215);
/*                 ": macro-to-key " */
        strcat(outseq, &bufn[1]);

        /* translate it to a buffer pointer */
        bufn[0] = '[';
        strcat(bufn, "]");
        if ((kmacro=bfind(bufn, FALSE, 0)) == NULL) {
                mlwrite(TEXT130);
/*              "Macro not defined"*/
                return(FALSE);
        }

        strcat(outseq, " ");
        mlwrite(outseq);

        /* get the command sequence to bind */
        c = getckey(FALSE);

        /* change it to something we can print as well */
        /* and dump it out */
        ostring(cmdstr(c, &outseq[0]));

#if QDOS & MENUS
        if (!strcmp(outseq, "M-X") || !strcmp(outseq, "^G") ||
            !strcmp(outseq, "^M")  || !strcmp(outseq, "^[")) {
                mlwrite(TEXT401);
/*                     "Error: This binding is reserved for the menu system." */
                return(FALSE);
        }
#endif
        /* search the table to see if it exists */
        ktp = &keytab[0];
        found = FALSE;
        while (ktp->k_type != BINDNUL) {
                if (ktp->k_code == c) {
                        found = TRUE;
                        break;
                }
                ++ktp;
        }

        if (found) {    /* it exists, just change it then */
                ktp->k_ptr.buf = kmacro;
                ktp->k_type = BINDBUF;
        } else {        /* otherwise we need to add it to the end */
                /* if we run out of binding room, bitch */
                if (ktp >= &keytab[NBINDS]) {
                        mlwrite(TEXT17);
/*                              "Binding table FULL!" */
                        return(FALSE);
                }

                ktp->k_code = c;        /* add keycode */
                ktp->k_ptr.buf = kmacro;        /* and the function pointer */
                ktp->k_type = BINDBUF;  /* and the binding type */
                ++ktp;                  /* and make sure the next is null */
                ktp->k_code = 0;
                ktp->k_type = BINDNUL;
                ktp->k_ptr.fp = NULL;
        }

        return(TRUE);
}

/* unbindkey:   delete a key from the key binding table */
#if PROTO
int PASCAL NEAR unbindkey(int f, int n)
#else
int PASCAL NEAR unbindkey(f, n)
int f, n;       /* command arguments [IGNORED] */
#endif
{
        register int c;         /* command key to unbind */
        char outseq[80];        /* output buffer for keystroke sequence */

        /* prompt the user to type in a key to unbind */
        mlwrite(TEXT18);
/*              ": unbind-key " */

        /* get the command sequence to unbind */
        c = getckey(FALSE);             /* get a command sequence */

        /* change it to something we can print as well */
        /* and dump it out */
        ostring(cmdstr(c, &outseq[0]));

#if QDOS & MENUS
        if (!strcmp(outseq, "M-X") || !strcmp(outseq, "^G") ||
            !strcmp(outseq, "^M")  || !strcmp(outseq, "^[")) {
                mlwrite(TEXT401);
/*                     "Error: This binding is reserved for the menu system." */
                return(FALSE);
        }
#endif

        /* if it isn't bound, bitch */
        if (unbindchar(c) == FALSE) {
                mlwrite(TEXT19);
/*                      "[Key not bound]" */
                return(FALSE);
        }
        return(TRUE);
}

#if PROTO
int PASCAL NEAR unbindchar(int c)
#else
int PASCAL NEAR unbindchar(c)
int c;          /* command key to unbind */
#endif
{
        register KEYTAB *ktp;   /* pointer into the command table */
        register KEYTAB *sktp;  /* saved pointer into the command table */
        register int found;     /* matched command flag */

        /* search the table to see if the key exists */
        ktp = &keytab[0];
        found = FALSE;
        while (ktp->k_type != BINDNUL) {
                if (ktp->k_code == c) {
                        found = TRUE;
                        break;
                }
                ++ktp;
        }

        /* if it isn't bound, bitch */
        if (!found)
                return(FALSE);

        /* save the pointer and scan to the end of the table */
        sktp = ktp;
        while (ktp->k_type != BINDNUL)
                ++ktp;
        --ktp;          /* backup to the last legit entry */

        /* copy the last entry to the current one */
        sktp->k_code = ktp->k_code;
        sktp->k_type = ktp->k_type;
        if (sktp->k_type == BINDFNC)
                sktp->k_ptr.fp   = ktp->k_ptr.fp;
        else if (sktp->k_type == BINDBUF)
                sktp->k_ptr.buf   = ktp->k_ptr.buf;

        /* null out the last one */
        ktp->k_code = 0;
        ktp->k_type = BINDNUL;
        ktp->k_ptr.fp = NULL;
        return(TRUE);
}

/* unbind all the keys bound to a buffer (which we can then delete */
#if PROTO
VOID PASCAL NEAR unbind_buf(BUFFER *bp)
#else
VOID PASCAL NEAR unbind_buf(bp)
BUFFER *bp;     /* buffer to unbind all keys connected to */
#endif
{
        register KEYTAB *ktp;   /* pointer into the command table */

        /* search the table to see if the key exists */
        ktp = &keytab[0];
        while (ktp->k_type != BINDNUL) {
                if (ktp->k_type == BINDBUF) {
                        if (ktp->k_ptr.buf == bp) {
                                unbindchar(ktp->k_code);
                                --ktp;
                        }
                }
                ++ktp;
        }
}

/* Describe bindings:

           bring up a fake buffer and list the key bindings
           into it with view mode
*/
#if PROTO
int PASCAL NEAR desbind(int f, int n)
#else
int PASCAL NEAR desbind(f, n)
int f,n;        /* prefix flag and argument */
#endif
{
        return(buildlist(TRUE, ""));
}

/* Apropos (List functions that match a substring) */
#if PROTO
int PASCAL NEAR apro(int f, int n)
#else
int PASCAL NEAR apro(f, n)
int f,n;        /* prefix flag and argument */
#endif
{
        char mstring[NSTRING];  /* string to match cmd names to */
        int status;             /* status return */

        status = mlreply(TEXT20, mstring, NSTRING - 1);
/*                       "Apropos string: " */
        if (status != TRUE)
                return(status);

        return(buildlist(FALSE, mstring));
}

/* build a binding list (limited or full) */
#if PROTO
int PASCAL NEAR buildlist(int type, char *mstring)
#else
int PASCAL NEAR buildlist(int type, char *mstring)
int type;       /* true = full list,   false = partial list */
char *mstring;  /* match string if a partial list */
#endif
{
        register KEYTAB *ktp;   /* pointer into the command table */
        register NBIND *nptr;   /* pointer into the name binding table */
        register BUFFER *listbuf;/* buffer to put binding list into */
        register BUFFER *bp;    /* buffer ptr for function scan */
        int cpos;               /* current position to use in outseq */
        char outseq[80];        /* output buffer for keystroke sequence */
        int first_entry;        /* is this the first macro listing? */
        
        /* get a buffer for the binding list */
        listbuf = bfind(TEXT21, TRUE, BFINVS);
/*                 "Binding list" */
        if (listbuf == NULL || bclear(listbuf) == FALSE) {
                mlwrite(TEXT22);
/*                      "Can not display binding list" */
                return(FALSE);
        }

        /* let us know this is in progress */
        mlwrite(TEXT23);
/*              "[Building binding list]" */

        /* build the contents of this window, inserting it line by line */
        nptr = &names[0];
        while (nptr->n_func != NULL) {

                /* add in the command name */
                strcpy(outseq, nptr->n_name);
                cpos = strlen(outseq);

                /* if we are executing an apropos command..... */
                if (type == FALSE &&
                    /* and current string doesn't include the search string */
                    strinc(outseq, mstring) == FALSE)
                        goto fail;

                /* search down any keys bound to this */
                ktp = &keytab[0];
                while (ktp->k_type != BINDNUL) {
                        if (ktp->k_type == BINDFNC &&
                            ktp->k_ptr.fp == nptr->n_func) {
                                /* padd out some spaces */
                                while (cpos < 25)
                                        outseq[cpos++] = ' ';

                                /* add in the command sequence */
                                cmdstr(ktp->k_code, &outseq[cpos]);

                                /* and add it as a line into the buffer */
                                if (addline(listbuf, outseq) != TRUE)
                                        return(FALSE);

                                cpos = 0;       /* and clear the line */
                        }
                        ++ktp;
                }

                /* if no key was bound, we need to dump it anyway */
                if (cpos > 0) {
                        outseq[cpos] = 0;
                        if (addline(listbuf, outseq) != TRUE)
                                return(FALSE);
                }

fail:           /* and on to the next name */
                ++nptr;
        }

        /* scan all buffers looking for macroes and their bindings */
        first_entry = TRUE;
        bp = bheadp;
        while (bp) {

                /* is this buffer a macro? */
                if (bp->b_bname[0] != '[')
                        goto bfail;

                /* add in the command name */
                strcpy(outseq, bp->b_bname);
                cpos = strlen(outseq);

                /* if we are executing an apropos command..... */
                if (type == FALSE &&
                    /* and current string doesn't include the search string */
                    strinc(outseq, mstring) == FALSE)
                        goto bfail;

                /* search down any keys bound to this macro */
                ktp = &keytab[0];
                while (ktp->k_type != BINDNUL) {
                        if (ktp->k_type == BINDBUF &&
                            ktp->k_ptr.buf == bp) {
                                /* padd out some spaces */
                                while (cpos < 25)
                                        outseq[cpos++] = ' ';

                                /* add in the command sequence */
                                cmdstr(ktp->k_code, &outseq[cpos]);

                                /* and add it as a line into the buffer */
                                if (addline(listbuf, outseq) != TRUE)
                                        return(FALSE);

                                cpos = 0;       /* and clear the line */
                        }
                        ++ktp;
                }

                /* add a blank line between the key and macro lists */
                if (first_entry == TRUE) {
                        if (addline(listbuf, "") != TRUE)
                                return(FALSE);
                        first_entry = FALSE;
                }

                /* if no key was bound, we need to dump it anyway */
                if (cpos > 0) {
                        outseq[cpos] = 0;
                        if (addline(listbuf, outseq) != TRUE)
                                return(FALSE);
                }

bfail:          /* and on to the next buffer */
                bp = bp->b_bufp;
        }

        wpopup(listbuf);
        mlerase();      /* clear the mode line */
        return(TRUE);
}

/* does source include sub? */
#if PROTO
int PASCAL NEAR strinc(char *source, char *sub)
#else
int PASCAL NEAR strinc(source, sub)
char *source;   /* string to search in */
char *sub;      /* substring to look for */
#endif
{
        char *sp;       /* ptr into source */
        char *nxtsp;    /* next ptr into source */
        char *tp;       /* ptr into substring */

        /* for each character in the source string */
        sp = source;
        while (*sp) {
                tp = sub;
                nxtsp = sp;

                /* is the substring here? */
                while (*tp) {
                        if (*nxtsp++ != *tp)
                                break;
                        else
                                tp++;
                }

                /* yes, return a success */
                if (*tp == 0)
                        return(TRUE);

                /* no, onward */
                sp++;
        }
        return(FALSE);
}

/* get a command key sequence from the keyboard */
#if PROTO
unsigned int PASCAL NEAR getckey(int mflag)
#else
unsigned int PASCAL NEAR getckey(mflag)
int mflag;      /* going for a meta sequence? */
#endif
{
        register unsigned int c;        /* character fetched */
        char tok[NSTRING];              /* command incoming */

        /* check to see if we are executing a command line */
        if (clexec) {
                macarg(tok);    /* get the next token */
                return(stock((unsigned char *)tok));
        }

        /* or the normal way */
        if (mflag)
                c = get_key();
        else
                c = getcmd();
        return(c);
}

/* execute the startup file */
#if PROTO
int PASCAL NEAR startup(char *sfname)
#else
int PASCAL NEAR startup(sfname)
char *sfname;   /* name of startup file (null if default) */
#endif
{
        char *fname;    /* resulting file name to execute */
        char name[NSTRING];     /* name with extention */

        /* look up the startup file */
        if (*sfname != 0) {

                /* default the extention */
                strcpy(name, sfname);
#if QDOS
                if (sindex(name, ".") == 0 && sindex(name, "_") == 0)
                        strcat(fname, "_cmd");
#else
                if (sindex(name, ".") == 0)
                        strcat(name, ".cmd");
#endif

                fname = flook(name, TRUE);
        } else
                fname = flook((char *)pathname[0], TRUE);

        /* if it isn't around, don't sweat it */
        if (fname == NULL)
                return(TRUE);

        /* otherwise, execute the sucker */
        return(dofile(fname));
}

/*      Look up the existance of a file along the normal or PATH
        environment variable.

        LOOKUP ORDER:

                if contains path:

                        absolute

                else

                        HOME environment directory
                        all directories along PATH environment
                        directories in table from EPATH.H

        NOTE for QDOS users:-

        If you want to find things in the PROG_USE directory with this
        function (e.g. when loading command files or executing programs)
        then you must add the program directory to the PATH environment
        variable, with (under QDOS) is a series of directory names
        separated by semicolons, for example:-

                flp1_;flp1_utils_;flp2_;flp2_utils_

        Note that there is NOT a terminating semicolon. The HOME
        environment variable mentioned above is treated by the following
        as the DATA_USE directory. If PATH is undefined the PROG_USE is
        used to replace it - normally it would not contain any
        semicolons in this case.

        For tidiness, directory names should always end in an underscore
        as in the example above.
*/
#if PROTO
char *PASCAL NEAR flook(char *fname, int hflag)
#else
char *PASCAL NEAR flook(fname, hflag)
char *fname;    /* base file name to search for */
int hflag;      /* Look in the HOME environment variable first? */
#endif
{
        register char *home;    /* path to home directory */
        register char *path;    /* environmental PATH variable */
        register char *sp;      /* pointer into path spec */
        register int i;         /* index */
        static char fspec[NFILEN];      /* full path spec to search */

        /* if we have an absolute path.. check only there! */
#if QDOS
        int extra_info;
        /*
         * QDOS does same thing a different way
         */
        if(isdevice(fname, &extra_info)) {
                if(ffropen(fname) == FIOSUC) {
                        ffclose();
                        return(fname);
                } else
                        return(NULL);
        }
#else
        sp = fname;
        while (*sp) {
                if (*sp == ':' || *sp == '\\' || *sp == '/') {
                        if (ffropen(fname) == FIOSUC) {
                                ffclose();
                                return(fname);
                        } else
                                return(NULL);
                }
                ++sp;
        }
#endif

#if     ENVFUNC

        if (hflag) {
#if WMCS
                home = getenv("SYS$HOME");
#elif QDOS
                home = getenv("EMACS_HOME");
                if (home == NULL ||
                    (strcmp(fname, pathname[0])
                    && strcmp(fname, "uelocal_rc")
                    && strcmp(fname, pathname[1])
                    && (strstr(fname, "_cmd") != fname+strlen(fname)-4)
                    && (strstr(fname, ".cmd") != fname+strlen(fname)-4)
                    && (strstr(fname, "ehelp1") != fname)
                    && (strstr(fname, "ehelp2") != fname)))
                        home = getenv("DATA_USE");
#else
                home = getenv("HOME");
#endif
                if (home != NULL) {
                        /* build home dir file spec */
                        strcpy(fspec, home);
#if WMCS
                        strcat(fspec,fname);
#elif QDOS
                        if(fspec[strlen(fspec)-1] != DIRSEPCHAR)
                                strcat(fspec, DIRSEPSTR);
                        strcat(fspec, fname);
#else
                        strcat(fspec, DIRSEPSTR);
                        strcat(fspec, fname);
#endif

                        /* and try it out */
                        if (ffropen(fspec) == FIOSUC) {
                                ffclose();
                                return(fspec);
                        }
                }
        }
#endif

        /* current directory now overides everything except HOME var */
        if (ffropen(fname) == FIOSUC) {
                ffclose();
                return(fname);
        }

#if     ENVFUNC
        /* get the PATH variable */
#if WMCS
        path = getenv("OPT$PATH");
#else
#if OS2
        path = getenv("DPATH");
#else
        path = getenv("PATH");
#endif
#endif

#if QDOS
        if (path == NULL)
                path = getenv("PROG_USE");
#endif

        if (path != NULL)
                while (*path) {

                        /* build next possible file spec */
                        sp = fspec;
#if TOS || QDOS
                        while (*path && (*path != PATHCHR) && (*path != ','))
#else
                        while (*path && (*path != PATHCHR))
#endif
                                *sp++ = *path++;

                        /* add a terminating dir separator if we need it */
                        if ((sp != fspec) && (*(sp-1) != DIRSEPCHAR))
                                *sp++ = DIRSEPCHAR;
                        *sp = 0;
                        strcat(fspec, fname);

                        /* and try it out */
                        if (ffropen(fspec) == FIOSUC) {
                                ffclose();
                                return(fspec);
                        }

#if     (TOS && MWC) || QDOS
                        if ((*path == PATHCHR) || (*path == ','))
#else
                        if (*path == PATHCHR)
#endif
                                ++path;
                }
#endif

        /* look it up via the old table method */
        for (i=2; i < NPNAMES; i++) {
                strcpy(fspec, pathname[i]);
                strcat(fspec, fname);

                /* and try it out */
                if (ffropen(fspec) == FIOSUC) {
                        ffclose();
                        return(fspec);
                }
        }

        return(NULL);   /* no such luck */
}

/* Change a key command to a string we can print out.
 * Return the string passed in.
 */
#if PROTO
char *PASCAL NEAR cmdstr(int c, char *seq)
#else
char *PASCAL NEAR cmdstr(c, seq)
int c;          /* sequence to translate */
char *seq;      /* destination string for sequence */
#endif
{
        char *ptr;      /* pointer into current position in sequence */

        ptr = seq;

        /* apply ^X sequence if needed */
        if (c & CTLX) {
                *ptr++ = '^';
                *ptr++ = 'X';
        }

        /* apply ALT key sequence if needed */
        if (c & ALTD) {
                *ptr++ = 'A';
                *ptr++ = '-';
        }

        /* apply Shifted sequence if needed */
        if (c & SHFT) {
                *ptr++ = 'S';
                *ptr++ = '-';
        }

        /* apply MOUS sequence if needed */
        if (c & MOUS) {
                *ptr++ = 'M';
                *ptr++ = 'S';
        }

        /* apply meta sequence if needed */
        if (c & META) {
                *ptr++ = 'M';
                *ptr++ = '-';
        }

        /* apply SPEC sequence if needed */
        if (c & SPEC) {
                *ptr++ = 'F';
                *ptr++ = 'N';
        }

        /* apply control sequence if needed */
        if (c & CTRL) {

                /* non normal spaces look like @ */
                if (ptr == seq && ((c & 255) == ' '))
                        c = '@';

                *ptr++ = '^';
        }

        c = c & 255;    /* strip the prefixes */

        /* and output the final sequence */
        *ptr++ = c;
        *ptr = 0;       /* terminate the string */
        return (seq);
}

/*      This function looks a key binding up in the binding table       */
#if PROTO
KEYTAB *getbind(int c)
#else
KEYTAB *getbind(c)
register int c; /* key to find what is bound to it */
#endif
{
        register KEYTAB *ktp;

        /* scan through the binding table, looking for the key's entry */
        ktp = &keytab[0];
        while (ktp->k_type != BINDNUL) {
                if (ktp->k_code == c)
                        return(ktp);
                ++ktp;
        }

        /* no such binding */
        return((KEYTAB *)NULL);
}

/* getfname:    This function takes a ptr to KEYTAB entry and gets the name
                associated with it
*/
#if PROTO
char *PASCAL NEAR getfname(KEYTAB *key)
#else
char *PASCAL NEAR getfname(key)
KEYTAB *key;    /* key binding to return a name of */
#endif
{
        int (PASCAL NEAR *func)(); /* ptr to the requested function */
        register NBIND *nptr;   /* pointer into the name binding table */
        register BUFFER *bp;    /* ptr to buffer to test */
        register BUFFER *kbuf;  /* ptr to requested buffer */

        /* if this isn't a valid key, it has no name */
        if (key == NULL)
                return(NULL);

        /* skim through the binding table, looking for a match */
        if (key->k_type == BINDFNC) {
                func = key->k_ptr.fp;
                nptr = &names[0];
                while (nptr->n_func != NULL) {
                        if (nptr->n_func == func)
                                return(nptr->n_name);
                        ++nptr;
                }
                return(NULL);
        }

        /* skim through the buffer list looking for a match */
        if (key->k_type == BINDBUF) {
                kbuf = key->k_ptr.buf;
                bp = bheadp;
                while (bp) {
                        if (bp == kbuf)
                                return(bp->b_bname);
                        bp = bp->b_bufp;
                }
                return(NULL);
        }
        return(NULL);
}

/* fncmatch:    match fname to a function in the names table and return
                any match or NULL if none */
#if PROTO
#if ZTC || TURBO || IC || C68
int (PASCAL NEAR *PASCAL NEAR fncmatch(char *fname))(int, int)
#else
int (PASCAL NEAR *PASCAL NEAR fncmatch(char *fname))(VOID)
#endif
#else
int (PASCAL NEAR *PASCAL NEAR fncmatch(fname))()
char *fname;    /* name to attempt to match */
#endif
{
        int nval;

        if ((nval = binary(fname, namval, numfunc, NSTRING)) == -1)
                return(NULL);
        else
                return(names[nval].n_func);
}

#if PROTO
char *PASCAL NEAR namval(int index)
#else
char *PASCAL NEAR namval(index)
int index;      /* index of name to fetch out of the name table */
#endif
{
        return(names[index].n_name);
}

/*      stock()         String key name TO Command Key

        A key binding consists of one or more prefix functions followed by
        a keystroke.  Allowable prefixes must be in the following order:

        ^X      preceeding control-X
        A-      similtaneous ALT key (on PCs mainly)
        S-      shifted function key
        MS      mouse generated keystroke
        M-      Preceding META key
        FN      function key
        ^       control key

        Meta and ^X prefix of lower case letters are converted to upper
        case.  Real control characters are automatically converted to
        the ^A form.
*/

#if PROTO
unsigned int PASCAL NEAR stock(unsigned char *keyname)
#else
unsigned int PASCAL NEAR stock(keyname)
unsigned char *keyname; /* name of key to translate to Command key form */
#endif
{
        register unsigned int c;        /* key sequence to return */

        /* parse it up */
        c = 0;

        /* Do ^X prefix */
        if(*keyname == '^' && *(keyname+1) == 'X') {
                if(*(keyname+2) != 0) { /* Key is not bare ^X */
                    c |= CTLX;
                    keyname += 2;
                }
        }

        /* and the ALT key prefix */
        if (*keyname == 'A' && *(keyname+1) == '-') {
                c |= ALTD;
                keyname += 2;
        }

        /* and the SHIFTED prefix */
        if (*keyname == 'S' && *(keyname+1) == '-') {
                c |= SHFT;
                keyname += 2;
        }

        /* and the mouse (MOUS) prefix */
        if (*keyname == 'M' && *(keyname+1) == 'S') {
                c |= MOUS;
                keyname += 2;
        }

        /* then the META prefix */
        if (*keyname == 'M' && *(keyname+1) == '-') {
                c |= META;
                keyname += 2;
        }

        /* next the function prefix */
        if (*keyname == 'F' && *(keyname+1) == 'N') {
                c |= SPEC;
                keyname += 2;
        }

        /* a control char?  (NOT Always upper case anymore) */
        if (*keyname == '^' && *(keyname+1) != 0) {
                c |= CTRL;
                ++keyname;
                if (*keyname == '@')
                        *keyname = ' ';
        }

        /* A literal control character? (Boo, hiss) */
        if (*keyname < 32) {
                c |= CTRL;
                *keyname += '@';
        }

        /* make sure we are not lower case if used with ^X or M- */
        if(!(c & (MOUS|SPEC|ALTD|SHFT)))        /* If not a special key */
            if( c & (CTLX|META))                /* If is a prefix */
                uppercase((unsigned char *)keyname);            /* Then make sure it's upper case */

        /* the final sequence... */
        c |= *keyname;
        return(c);
}

/* string key name to binding name.... */
#if PROTO
char *PASCAL NEAR transbind(char *skey)
#else
char *PASCAL NEAR transbind(skey)
char *skey;     /* name of key to get binding for */
#endif
{
        char *bindname;

        bindname = getfname(getbind(stock((unsigned char *)skey)));
        if (bindname == NULL)
                bindname = (char *)errorm;

        return(bindname);
}

/* execute a function bound to a key */
#if PROTO
int PASCAL NEAR execkey(KEYTAB *key, int f, int n)
#else
int PASCAL NEAR execkey(key, f, n)
KEYTAB *key;    /* key to execute */
int f, n;       /* agruments to C function */
#endif
{
        register int status;    /* error return */

        if (key->k_type == BINDFNC) {
                undo_insert(OP_CMND, 1, obj);
                return((*(key->k_ptr.fp))(f, n));
        }
        if (key->k_type == BINDBUF) {
                while (n--) {
                        status = dobuf(key->k_ptr.buf);
                        if (status != TRUE)
                                return(status);
                }
        }
        return(TRUE);
}

/* set a KEYTAB to the given name of the given type */
#if PROTO
int set_key(KEYTAB *key, char *name)
#else
int set_key(key, name)
KEYTAB *key;            /* ptr to key to set */
char *name;             /* name of function or buffer */
#endif
{
        int (PASCAL NEAR *ktemp)();     /* temp function pointer to assign */
        register BUFFER *kmacro;        /* ptr to buffer of macro to bind to key */
        char bufn[NBUFN];               /* buffer to hold macro name */

        /* are we unbinding it? */
        if (*name == 0) {
                key->k_type = BINDNUL;
                return(TRUE);
        }

        /* bind to a built in function? */
        if ((ktemp = fncmatch(name)) != NULL) {
                key->k_ptr.fp = ktemp;
                key->k_type = BINDFNC;
                return(TRUE);
        }

        /* is it a procedure/macro? */
        strcpy(bufn, "[");
        strcat(bufn, name);
        strcat(bufn, "]");
        if ((kmacro=bfind(bufn, FALSE, 0)) != NULL) {
                key->k_ptr.buf = kmacro;
                key->k_type = BINDBUF;
                return(TRUE);
        }

        /* not anything we can bind to */
        mlwrite(TEXT16);
/*              "[No such function]" */
        return(FALSE);
}
