/*
 * QL.C: Terminal driver for MicroEMACS 3.11-4.00
 *      (c) 1991,92 Richard Kettlewell - (c) 1993-1999 Thierry Godefroy.
 *
 *              16/11/91        Adapted to QDOS from IBMPC.C
 *              18/02/92        Updated for C68 release 2
 *              20/03/92        Updated for EMACS 3.11
 *              21/03/92        Took out some stuff to write in assembly
 *
 *      Modifications and additions by Thierry Godefroy:
 *
 *              29/03/93        Thor XVI 16 colours mode added ($sres=THOR).
 *              03/04/93        - ttopen() & ttclose() moved from qdos_c.
 *                              - updates_emacs_vars() moved to qdos_c.
 *                              - in_init(), in_check(), in_put() & in_get()
 *                                borrowed from msdos_c (to implement mouse
 *                                events).
 *                              - ttopen() & qlgetc() modified to implement
 *                                mouse events.
 *                              - checkmouse() added.
 *                              - pointer environment calls are now done
 *                                with libqptr v2.0 functions.
 *              16/04/93        Mouse events fully implemented.
 *              23/06/93        Pointer driven menus implemented (see also the
 *                               "menus_asm" file).
 *              06/01/94        Input buffer size increased for file names
 *                               returned by QMenu file selector.
 *              05/06/94        Cursor bug on QXL corrected.
 *              28/08/94        Type corrected for qlout (chanid_t), menubar()
 *                               and menu_job_id (jobid_t).
 *              27/08/95        Some tidying to remove compilation warnings.
 *              04/01/98        Made MOUSE and MENUS support independent.
 *              03/12/98        Adapted to MicroEMACS v4.00.
 *                              Menu bar removed, new main menu added.
 *              04/12/98        Menu resizing implemented.
 *              08/12/98        "About" box implemented, new pointers for appli-
 *                               cation window and main menu added.
 *              12/12/98        Changed main menu size so that it can fit a
 *                               standard 512x256 QL screen.
 *                              Changed move and resize routines so that no such
 *                               operation is attempted when the screen is only
 *                               512x256 large.
 *              13/12/98        Improved mouse events handling functions
 *                               (in checkmouse() and qlgetc()).
 *                              <ENTER> is now translated into ^M and <SHIFT
 *                               ENTER> into ^J. This is done in qlgetc().
 *                              Made the application window pointer configurable
 *                               (for Jonathan Hudson).
 *              14/12/98        Implemented page up/down and help icons in the
 *                               main menu.
 *                              A "do" on the resize icon now enlarges the menu
 *                               to the maximal size (screen size).
 *              15/12/98        The current buffer filename is now passed to the
 *                               pull-down menu routines so that FSEL can use it
 *                               for default directory and filename.
 *              16/12/98        Prototyping of all functions.
 *              20/12/98        Changed some cursor keys bindings and added the
 *                               "toggle OVER mode" binding to the <INS> key
 *                               (see toggleover() function in "random_c").
 *              26/12/98        Added the qlspeed() function to the terminal
 *                               driver (allows for a processor speed indepen-
 *                               dant pause in CMODE fence matching).
 *              29/12/98        The pointer no more disappears when moving it
 *                               from the main menu to the application sub-
 *                               window (i.e. a mouse event is read instead of
 *                               a keypress).
 *              01/01/99        Added the context sub-menu call in checkmouse().
 *              06/01/99        Improved menu_resize() function (the menu does
 *                               not move randomly anymore when resized). The
 *                               call to this function is now done from ASIZE()
 *                               instead of checkmouse().
 *              07/01/99        Added the EMACS_LINES and EMACS_COLS default
 *                               menu sizing scheme in ttopen().
 *                              New "wman" flag added so that MicroEMACS can now
 *                               work with PTR_GEN (pointer) but without WMAN
 *                               (menus).
 *                              Changed the cursor key bindings for "hunt
 *                               forward/backwards" so that <CTRL DOWN> is now
 *                               available for the "emacs_rc" macro
 *                               "kill-current-line".
 *              28/02/99        Added a work around for UQLX directory open bug
 *                               (enabled when "EMACS_UQLXBUG" is defined).
 *              13/05/99        Changed in_check() to implement CSM server
 *                               request automatic polling.
 *              03/11/99        Added support for more "colors" (in fact
 *                               patterns): DRED, DGREEN and GOLD (see also
 *                               edef_h and estruct_h).
 *                              A patch level is now appended to the MicroEMACS
 *                               version number shown in the "about" box.
 *                              Added the "fist" sprite (for dragging events in
 *                               the application sub-window).
 *              18/11/99        Character attributes support implemented: added
 *                               bold and italic fonts; added qlbold(),
 *                               qlitalic() and qlunder() pointers into the
 *                               "term" structure; added the qlboldstate and
 *                               qlitalstate flags.
 *              27/08/00        Added the "UNDO_KEY" conditional compilation for
 *                               bindings (binds the UNDO key of Ataris and the
 *                               PAUSE key of QXL+SMSQ/E v2.98+ to "undo last
 *                               operation").
 *              12/09/00        Added the scroll bar drawing and updating (it is
 *                               not yet "clickable" nor "draggable" though).
 *                              The page/line up/down loose items are no more
 *                               making the screen to flicker when used (a full
 *                               screen update is no more necessary).
 *              13/09/00        Added scroll bar moving/dragging support.
 *		20/09/00	Added the compiler name into the "about" box.
 *              24/09/00        Don't ignore any more a keypress occuring while
 *                               the pointer is active into the application sub-
 *                               window.
 *                              <SHIFT> <TAB> now allows to move the pointer
 *                               forth and back between the main menu and the
 *                               application sub-window.
 */

#if QDOS

#include        <stdio.h>
#include        <stdlib.h>
#include        <string.h>
#include        <qdos.h>
#include        <qptr.h>
#define  termdef 1                      /* don't define "term" external */
#include        "estruct.h"
#include        "eproto.h"
#include        "edef.h"
#if CHRATTR
#include        "qlfonts.h"
#endif

#undef COMPILER
#ifdef __GNUC__
#define COMPILER "gcc"
#else
#define COMPILER "c68"
#endif

#define UNDO_KEY                        /* Implement the <UNDO> key binding */

#define MARGIN 8                        /* Size of minimum margin and... */
#define SCRSIZ 64                       /* scroll size for extended lines. */
#define NPAUSE 25                       /* Pause after a fence matching
                                           in CMODE before resuming typing;
                                           defined in qdos_c too */

#define INFINITE (timeout_t)-1          /* Infinite timeout for QDOS calls */

/* Function defined in "qdos_c" */
void update_emacs_vars(void);

/* Functions defined in "fast_s" */

int qlmove(int row, int col), qleeol(void), qlputc(int ch), qleeop(void),
    qlrev(int state), qlopen(void), qlclose(void), qlflush(void),
    qlfcol(int color), qlbcol(int color), qlkopen(void), qlkclose(void),
    qlmouse(void), qlunder(int flag), qlbold(int flag), qlitalic(int flag);

/* Forward function declarations */

int qlgetc(void), qlcres(char *res), qlbeep(void), qlsleep(int pause);

/* Global variables */

int ctrl_row    = 2;                /* Code returned by KEYROW(7) for <CTRL> */

chanid_t qlout;                     /* Window channel ID */

extern char server_started;         /* CSM server flag, defined in qdos_c */

static char ptrenv = FALSE;         /* Flag (pointer environment) */
static char wman   = FALSE;         /* Flag (window manager) */
static struct QLRECT wsize;         /* Current window size in characters */
static struct QLRECT max_pixsize;   /* Screen size in pixels */

#if MOUSE
static char hidden_cursor = TRUE;   /* Flag: cursor hidden by checkmouse() ? */
static int mousepos;
static int mousebut         = 0;
static struct WM_prec mouse;
#endif

#if MENUS
/* Function defined in "display_c" */
int     winch_vtresize(int rows, int cols);

/* Function declared in "menubutton_c" */
void    menu_button(struct WM_wwork *wwk, char *title);

/* Function declared in "menus_asm" */
void    killer(int flag);
jobid_t pop_menu(int menu_number, char * fname);

/* Global variables */
static int lit_number;
static short delta_x, delta_y;
int appw_pointer, win_lines, win_cols;

/* Main menu definition */
#define ninfw  3    /* number of information windows */
#define ninfo  1    /* number of information objects */
#define nlitm 16    /* number of loose items */
#define nappl  1    /* number of application sub-windows */

static long AMOVE    (struct WM_wwork *, struct WM_litm *);
static long ASIZE    (struct WM_wwork *, struct WM_litm *);
static long ASLEEP   (struct WM_wwork *, struct WM_litm *);
static long lit_event(struct WM_wwork *, struct WM_litm *);
static long AHIT     (struct WM_wwork *, struct WM_appw *,
                      struct WM_wstat *, short, short, short);
static long ADRAW    (struct WM_wwork *, struct WM_appw *);
static long ACTRL    (struct WM_wwork *, struct WM_appw *,
                      struct WM_wstat *, short, short, short, short);

static struct WM_action amove  = {JSR, wm_actli, AMOVE};
static struct WM_action asize  = {JSR, wm_actli, ASIZE};
static struct WM_action asleep = {JSR, wm_actli, ASLEEP};
static struct WM_action afun   = {JSR, wm_actli, lit_event};
static struct WM_action ahit   = {JSR, wm_hitaw, AHIT};
static struct WM_action adraw  = {JSR, wm_drwaw, ADRAW};
static struct WM_action actrl  = {JSR, wm_ctlaw, ACTRL};

static void menu_resize(void);
static void about(void);

/* Window status area */

static struct WM_wstat ws = {
     NULL,                  /* *wwork               point back to work def */
     NULL,                  /* *wdef                this was wdef */ 
     0,                     /* chid                 chan ID enclosing pointer */ 
     0,0,0,                 /* swnr, xpos, ypos     sub-window and ptr posn */ 
     0,0,                   /* kstk, kprs           key stroke and press */ 
     0,                     /* evnt                 event vector */ 
     0,0,0,0,               /* xsiz, ysiz, xorg, yorg  */ 
     0,0,                   /* ptpsx, ptpsy         pointer position (abs) */ 
     FORM_QL4,0,0,          /* wmode, 2 spare       window mode and spare */
     NULL,                  /* *ciact               current item "action" */ 
     -1,                    /* citem                current item */ 
     0,0,0,                 /* cibrw, cipap, cispr  current item attr */ 
     0,0,0,0,               /* cihxs,ys, cihxo,yo   current item hit area */ 
     {0,0,0,0,0,0,0,0}      /* litem[40]            loose item status */ 
};

/* Text objects */

static QD_TEXTI(ob1_mainmenu,PROGNAME " v" VERSION);
static QD_TEXTI(ob2_mainmenu,"Files");
static QD_TEXTI(ob3_mainmenu,"Edit");
static QD_TEXTI(ob4_mainmenu,"Regions");
static QD_TEXTI(ob5_mainmenu,"Search");
static QD_TEXTI(ob6_mainmenu,"Moves");
static QD_TEXTI(ob7_mainmenu,"Buff./Screens");
static QD_TEXTI(ob8_mainmenu,"Windows");
static QD_TEXTI(ob9_mainmenu,"Var./Modes");
static QD_TEXTI(ob10_mainmenu,"Commands");
static QD_TEXTI(ob11_mainmenu,"?");
static QD_TEXTI(ob12_mainmenu,"Zzz");

/* Sprites */

extern WM_sprite_t wm_wprite_move, wm_sprite_size,
                   info_spr, pgup_spr, pgdn_spr,
                   curs_spr, hand_spr, fist_spr;

/* Information objects */

static struct WM_info ol1_mainmenu[ninfo+1] = {
    { 98, 10,               /* xsize, ysize      size */
       8,  1,               /* xorg, yorg        and origin */
       0,  0,               /* type, spare       item type */
      {7, 0, 0},            /* ink, cwid, chgt   character attributes */
      &ob1_mainmenu},       /* *pobj             pointer to object */

    {-1}
};

/* Information windows */

static struct WM_infw infw[ninfw+1] = {
    {416, 14,               /* xsize, ysize  */
      46,  0,               /* xorg, yorg */
       0,                   /* flag */
       0,  0,100,           /* Borw, Borc,papc    character attributes */
     NULL},                 /* *pobj             pointer to object */

    {114, 11,               /* xsize, ysize  */
     194,  2,               /* xorg, yorg */
       0,                   /* flag */
       0,  0,  0,           /* Borw, Borc,papc    character attributes */
     ol1_mainmenu},         /* *pobj             pointer to object */

    {500, 13,               /* xsize, ysize  */
       4, 16,               /* xorg, yorg */
       0,                   /* flag */
       1,  2,  0,           /* Borw, Borc,papc    character attributes */
     NULL},                 /* *pobj             pointer to object */

   {-1}
};

/* Loose items */

static struct WM_litm litm[nlitm+1] = {
    { 36, 11,               /* xsize, ysize  */
       6, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
      -1,'F',               /* Type ,Skey */
     &ob2_mainmenu,         /* Pointer to object */
       0,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 30, 11,               /* xsize, ysize  */
      44, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
      -1,'E',               /* Type ,Skey */
     &ob3_mainmenu,         /* Pointer to object */
       1,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 48, 11,               /* xsize, ysize  */
      76, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
      -1,'R',               /* Type ,Skey */
     &ob4_mainmenu,         /* Pointer to object */
       2,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 42, 11,               /* xsize, ysize  */
     126, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
      -1,'S',               /* Type ,Skey */
     &ob5_mainmenu,         /* Pointer to object */
       3,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 36 ,11,               /* xsize, ysize  */
     170, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
      -1,'M',               /* Type ,Skey */
     &ob6_mainmenu,         /* Pointer to object */
       4,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 82, 11,               /* xsize, ysize  */
     208, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
      -1,'B',               /* Type ,Skey */
     &ob7_mainmenu,         /* Pointer to object */
       5,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 48, 11,               /* xsize, ysize  */
     292, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
      -1,'W',               /* Type ,Skey */
     &ob8_mainmenu,         /* Pointer to object */
       6,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 64, 11,               /* xsize, ysize  */
     342, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
      -1,'V',               /* Type ,Skey */
     &ob9_mainmenu,         /* Pointer to object */
       7,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 52, 11,               /* xsize, ysize  */
     408, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
      -1,'C',               /* Type ,Skey */
     &ob10_mainmenu,        /* Pointer to object */
       8,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    {  8, 11,               /* xsize, ysize  */
     462, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
       0,  4,               /* Type,Skey */
     &ob11_mainmenu,        /* Pointer to object */
       9,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 20, 13,               /* xsize, ysize  */
      24,  1,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
       0,  7,               /* Type,Skey */
     &ob12_mainmenu,        /* Pointer to object */
      10,                   /* Item Number */
     &asleep},              /* *pact   pointer to action */

    { 20, 13,               /* xsize, ysize  */
       2,  1,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
       2,  5,               /* Type,Skey */
     &wm_sprite_move,       /* Pointer to object */
      11,                   /* Item Number */
     &amove},               /* *pact   pointer to action */

    { 20, 13,               /* xsize, ysize  */
     486,  1,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
       2,  6,               /* Type,Skey */
     &wm_sprite_size,       /* Pointer to object */
      12,                   /* Item Number */
     &asize},               /* *pact   pointer to action */

    { 20, 13,               /* xsize, ysize  */
     464,  1,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
       2,236,               /* Type,Skey */
     &info_spr,             /* Pointer to object */
      13,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 12, 11,               /* xsize, ysize  */
     472, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
       2,217,               /* Type,Skey */
     &pgdn_spr,             /* Pointer to object */
      14,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

    { 12, 11,               /* xsize, ysize  */
     486, 17,               /* xorg, yorg */
       0,  0,               /* xjst,yjst */
       2,209,               /* Type,Skey */
     &pgup_spr,             /* Pointer to object */
      15,                   /* Item Number */
     &afun},                /* *pact   pointer to action */

   {-1}
  };

/* Application sub-windows */

static WM_PWCB (1) ycb = {0};       /* Part window ctrl block for pan bar */

static struct WM_menw aw0_mainmenu = {
    492,222,                        /* xsize,ysize */
      4, 30,                        /* xorg,yorg */
    0,                              /* flag */
    1,2,0,                          /* borw,borc,papr */
    &curs_spr,                      /* *pspr Pointer to sprite  */
    &adraw,                         /* *draw routine */
    &ahit,                          /* *hit routine  */
    &actrl,                         /* *ctrl routine */                   
    0,1,                            /* nxsc , nysc   Max x/y sections */
    (char)253,                      /* Selection key press (SHIFT TAB) */
    0,0,                            /* 2 spare */
    NULL,                           /* *pwcb Part window control block */
    0,0,0,0,0,0,NULL,NULL,          /* 6 short 2 pointer zero */
    CL_WRROW,CL_WBARB,CL_WBARS,     /* pasc psbc pssc arrow/bar clrs */
    (WM_pwcb_t *) &ycb,		    /* *pwcb Part window ctrl block */
    0,0,0,0,0,0,NULL,NULL,          /* 6 short 2 pointer zero */
    7,7,7,                          /* pasc psbc pssc arrow/bar clrs */
    NULL                            /* pointer to status area */
};

static struct WM_appl appl[nappl+1] = {(WM_appw_t *) &aw0_mainmenu, NULL};

/* Window working definition */

static struct WM_wwork ww = {
    &ws,                    /* *wstat;              pointer to status area */
    NULL,                   /* *wscale              this was pointer to WDEF */
    0,                      /* chid                 channel ID */
    NULL,                   /* *pprec               pointer to pointer record */
    0,                      /* psave                pointer position saved */
    0,0,0,0,                /* 4 zero */
    appl,                   /* *splst               pointer to sprite list */
    508,254,                /* xsize, ysize         size */
     65,  5,                /* xorg, yorg           and origin */
    0,                      /* flag */
    1,7,0,                  /* borw, borc, papr     border width/clr, paper */
    &hand_spr,              /* *sprite              pointer to sprite */
    1,4,                    /* curw, curc           current item width/clr */
    0,2,NULL,NULL,          /* uback, uink, *ublob, *upatt  unavail */
    0,7,NULL,NULL,          /* aback, aink, *ablob, *apatt  avail */
    2,7,NULL,NULL,          /* sback, sink, *sblob, *spatt  selected */
    NULL,                   /* *help */
    ninfw, ninfo, infw,     /* ninfo, ninob, *pinfo info winds */
    nlitm, litm,            /* nlitm, *plitm        loose items */
    nappl, appl             /* nappl  *pappl        appl winds */
};

#endif /* MENUS */

/*
 * The treatment of colour in Ql EMACS has evolved rather.
 *
 * The variables cfcolour and cbcolour reflect the current logical ink and strip
 * colours. If qlrevstate == 0 then they are also the physical ink and strip 
 * colours; otherwise, they are reversed.
 *
 * Whenever qlfcol(), qlbcol() or qlrev() is called, the appropriate changes are
 * made and updates to the physical colours are made by calling QDOS.
 *
 * The paper colour (as distinct from the strip colour) is only set when a
 * 'clear' function is about to be called (i.e. in qleeop() and qleeol()).
 *
 * The former, qleeop(), always sets the paper colour to the global background
 * colour, since this is what it is expected to clear the screen too.
 *
 * The EMACS colours are translated to the logical colours by means of the
 * lookup array qlctrans. I have chosen to make the light colours stipples of
 * the named colour with white (this may change).
 *
 * All of the function relating to colour and reverse text are now in fast.s;
 * only the variables are here.
 *
 * In summary,
 *
 *      EMACS                   Logical                 Physical
 *      colours        -->      colours        -->      colours
 *                  qlctrans[]              qlrevstate
 */

char cfcolour = 7;                      /* Logical background colour */
char cbcolour = 0;                      /* Logical foreground colour */
char qlrevstate = 0;                    /* Reverse status */
char qlboldstate = 0;                   /* Bold status */
char qlitalstate = 0;                   /* Italic status */

unsigned char qlctrans[] = {            /* EMACS -> logical colour translation */
    0,                                  /* Black [this is startup paper colour] */
    2,                                  /* Red */
    4,                                  /* Green */
    6,                                  /* Yellow */
    1,                                  /* Blue */
    3,                                  /* Magenta */
    5,                                  /* Cyan */
    7,                                  /* "Grey" [this is startup text colour] */
    0|(7^0)<<3|3<<6,                    /* "Gray" */
    2|(7^2)<<3|3<<6,                    /* LRed */
    4|(7^4)<<3|3<<6,                    /* LGreen */
    6|(7^6)<<3|3<<6,                    /* LYellow */
    1|(7^1)<<3|3<<6,                    /* LBlue */
    3|(7^3)<<3|3<<6,                    /* LMagenta */
    5|(7^5)<<3|3<<6,                    /* LCyan */
    7,                                  /* White */
    208,                                /* DRed */
    232,                                /* DGreen */
    242                                 /* Gold */
};

/*
 * Since most micros have 10 (or even more) function keys, MicroEMACS
 * and many of its command files assume that there are indeed 10
 * function keys available. On the Ql, F1 to F5 are as normal; taken
 * with SHIFT, they represent F6 to F10 (i.e. FN6 to FN9 and FN0)
 * Taken with CTRL, all 10 combinations represent SHIFTed combinations
 * to EMACS. This is a bit of a kludge, but necessary because
 * some add-on keyboards for the Ql map F6 to F10 onto the shifted
 * F1 to F5.
 *
 * Other special keys are, with S- = SHIFT, ^ = CTRL, ~ = ALT,
 *
 *             Cursor back
 *      ~      Cursor to start of line
 *      ^      Delete back
 *      S-     Cursor back one word
 *      ^S-    Delete back one word
 *             Cursor forward
 *      ~      Cursor to end of line
 *      ^      Delete forward
 *      ^~     Kill to end of line (i.e. ^K)
 *      S-     Cursor forward one word
 *      ^S-    Delete forward one word
 *      ~S-    Insert space on the right
 *             Previous line
 *      ~      Previous page
 *      S-     Previous page
 *      ^S-    Move window one line up
 *      ^      Hunt backwards
 *      ~^     Undo last operation (if UNDO_KEY is defined, else go to start
 *              of paragraph)
 *      ~S-    Go to beginning of buffer
 *             Next line
 *      ~      Next page
 *      S-     Next page
 *      ^S-    Move window one line down
 *      ^      Hunt forwards
 *      ^~     Go to end of paragraph
 *      ~S-    Go to end of buffer
 *
 *      S-TAB   Same as TAB (if no MENUS)
 */

static short speclist[] = {         /* Table of special keys */
    SPEC|'B',                       /*     Cursor left */
    SPEC|ALTD|'B',                  /* ~   Start of line */
    0,                              /* ^   Hard binded to ^H */
    SPEC|CTRL|ALTD|'B',             /* ^~  */
    SPEC|CTRL|'B',                  /* S   Cursor left word */
    SPEC|ALTD|SHFT|'B',             /* ~S  */
    SPEC|CTRL|SHFT|'B',             /* ^S  Delete left word */
    SPEC|CTRL|ALTD|SHFT|'B',        /* ^~S */

    SPEC|'F',                       /*     Cursor right */
    SPEC|ALTD|'F',                  /* ~   End of line */
    SPEC|'D',                       /* ^   Delete right */
    SPEC|CTRL|ALTD|'F',             /* ^~  Kill to end of line*/
    SPEC|CTRL|'F',                  /* S   Cursor right word */
    CTRL|'C',                       /* ~S  Insert space */
    SPEC|CTRL|SHFT|'F',             /* ^S  Delete right word */
    SPEC|CTRL|ALTD|SHFT|'F',        /* ^~S */

    SPEC|'P',                       /*     Cursor up */
    SPEC|'Z',                       /* ~   Page up */
    SPEC|CTRL|'P',                  /* ^   */
#ifdef UNDO_KEY
    CTLX|CTRL|'H',                  /* ^~ (PAUSE for IBM kbd under QXL+SMSQ/E
                                     *      v2.98+ = UNDO for Atari kbd):
                                     * undo last operation
                                     */
#else
    SPEC|CTRL|'Z',                  /* ^~  Beginning of paragraph */
#endif
    SPEC|'Z',                       /* S   Page up */
    SPEC|'<',                       /* ~S  Beginning of buffer */
    CTLX|CTRL|'P',                  /* ^S  Move window one line up */
    ALTD|'R',                       /* ^~S Hunt backwards */

    SPEC|'N',                       /*     Cursor down */
    SPEC|'V',                       /* ~   Page down */
    SPEC|CTRL|'N',                  /* ^   */
    SPEC|CTRL|'V',                  /* ^~  End of paragraph */
    SPEC|'V',                       /* S   Page down */
    SPEC|'>',                       /* ~S  End of buffer */
    CTLX|CTRL|'N',                  /* ^S  Move window one line down */
    ALTD|'S',                       /* ^~S Hunt forwards */

    SPEC|'?',                       /* CAPS */
    SPEC|ALTD|'?',                  /* ~CAPS */
    SPEC|CTRL|'?',                  /* ^CAPS */
    SPEC|ALTD|CTRL|'?',             /* ^~CAPS */
    SPEC|SHFT|'?',                  /* S-CAPS */
    SPEC|SHFT|ALTD|'?',             /* ~S-CAPS */
    SPEC|SHFT|CTRL|'?',             /* ^S-CAPS */
    SPEC|SHFT|ALTD|CTRL|'?',        /* ^~S-CAPS */

    SPEC|'1',                       /* F1   */
    SPEC|SHFT|'1',                  /* ^F1  */
    SPEC|'6',                       /* F6   */
    SPEC|SHFT|'6',                  /* ^F6  */
    SPEC|'2',                       /* F2   */
    SPEC|SHFT|'2',                  /* ^F2  */
    SPEC|'7',                       /* F7   */
    SPEC|SHFT|'7',                  /* ^F7  */
    SPEC|'3',                       /* F3   */
    SPEC|SHFT|'3',                  /* ^F3  */
    SPEC|'8',                       /* F8   */
    SPEC|SHFT|'8',                  /* ^F8  */
    SPEC|'4',                       /* F4   */
    SPEC|SHFT|'4',                  /* ^F4  */
    SPEC|'9',                       /* F9   */
    SPEC|SHFT|'9',                  /* ^F9  */
    SPEC|'5',                       /* F5   */
    SPEC|SHFT|'5',                  /* ^F5  */
    SPEC|'0',                       /* F10  */
    SPEC|SHFT|'0',                  /* ^F10 */

    META|CTRL|'O',                  /* S-SPACE  Toggle OVER mode */
    '\t',                           /* S-TAB    Hard binded if MENUS */
};

/* Standard terminal interface dispatch table. */

TERM term = {
    0,0,0,0,                        /* Find screen size at run time */
    0,0,                            /* Origin is new in 3.11 */
    MARGIN, SCRSIZ, NPAUSE,
    qlopen, qlclose, qlkopen, qlkclose,
    qlgetc, qlputc, qlflush, qlmove,
    qleeol, qleeop, qleeop,
    qlbeep, qlrev, qlcres,
#if COLOR
    qlfcol, qlbcol,
#endif
#if CHRATTR
    qlunder, qlbold, qlitalic,
#endif
    qlsleep
};

/* Input buffers and pointers. */

#define IBUFSIZE 128                    /* this must be a power of 2 */

static unsigned char in_buf[IBUFSIZE];  /* input character buffer */
static int in_next = 0;                 /* pos to retrieve next input char */
static int in_last = 0;                 /* pos to put most recent input char */

/* Initialize the input buffer. */

void in_init(void)
{
    in_next = in_last = 0;
}

/* Is the input buffer non-empty ? */

int in_check(void)
{
    if (server_started == 1)                /* First check if the CSM server */
        pollserver(FALSE, 0);               /*  received a request */

    if (in_next == in_last)
        return FALSE;
    else
        return TRUE;
}

/* Put an event into the input buffer. */

void in_put(int event)
{
    in_buf[in_last++] = (unsigned char)event;
    in_last &= (IBUFSIZE - 1);
}

/* Get an event from the input buffer. */

int in_get(void)
{
    register int event;             /* event to return */

    event = (int)in_buf[in_next++];
    in_next &= (IBUFSIZE - 1);
    return event;
}

/* This function is called once to set up the terminal device streams. */

int ttopen(void)
{
    short  smode=-1, dtype=-1;
    long   error;

    /* Set up environment variable effects. */
    update_emacs_vars();

    /* Deactivate safe-save mode; you can reactivate it in a startup file.
     * Further, safe save mode has produced some distinctly odd results when used
     * relating to files not in the current directory (Nothing fatal, though).
     *
     * 6/9/1992 Made this conditional on EMACS_SSAVE
    */
    if (getenv("EMACS_SSAVE"))
        ssave = TRUE;
    else
        ssave = FALSE;

    in_init();                                  /* Initialize keyboard buffer */

    qlout = fgetchid(stdout);                   /* Find channel ID */

    /* IOP.FLIM call. On systems with Pointer Interface software installed, this
     * returns a sensible result; otherwise, an error is generated so we use the
     * "traditional" 512x256 screen size.
     */
    error = iop_flim(qlout, INFINITE, (WM_wsiz_t *) &max_pixsize);

    if (error) {                                /* If error, use default */
        max_pixsize.q_width  = 512;
        max_pixsize.q_height = 256;
        max_pixsize.q_x      = 0;
        max_pixsize.q_y      = 0;
        sd_wdef(qlout, INFINITE, 0, 0, &max_pixsize);   /* Set window size */
    } else {
        ptrenv = TRUE;                                  /* Ptr env is present */
#if MOUSE
#if MENUS
        if (wm_findv(qlout))
            wman = TRUE;
        if (wman) {
            ww.chid = qlout;
            if (appw_pointer)
                aw0_mainmenu.pspr = NULL;
            ws.wwork = &ww;
            if (6 * win_cols > aw0_mainmenu.xsize) {
                delta_x = 6 * win_cols + 8 - ww.xsize;
                if (ww.xsize + delta_x > max_pixsize.q_width - 4) {
                    delta_x = max_pixsize.q_width - ww.xsize - 4;
                    delta_x = 6 * (delta_x / 6);
                }
                ww.xsize += delta_x;
                aw0_mainmenu.xsize += delta_x;
                infw[0].xsize += delta_x;
                infw[2].xsize += delta_x;
                litm[12].xorg += delta_x;
                litm[13].xorg += delta_x;
                litm[14].xorg += delta_x;
                litm[15].xorg += delta_x;
                infw[1].xorg   = ((ww.xsize - infw[1].xsize - 2) >> 2) * 2;
            }
            if (10 * win_lines > aw0_mainmenu.ysize) {
                delta_y = ww.ysize - 10 * win_lines - 32;
                if (ww.ysize - delta_y > max_pixsize.q_height - 2) {
                    delta_y = ww.ysize - max_pixsize.q_height + 2;
                    delta_y = 10 * (delta_y / 10);
                }
                ww.ysize -= delta_y;
                aw0_mainmenu.ysize -= delta_y;
            }
            aw0_mainmenu.ypwcb->nsec = 1;
            aw0_mainmenu.nrow = aw0_mainmenu.ysize;
            aw0_mainmenu.ypwcb->s[0].siz = aw0_mainmenu.ysize;
            aw0_mainmenu.ypwcb->s[0].stt = 1;
            wm_prpos(&ww, -1, -1);
            wm_wdraw(&ww);
            wm_swapp(&ww, 0, 7);
        } else {
            /* Set window outline and size */
            iop_outl(qlout, INFINITE, 0, 0, 0, (WM_wsiz_t *) &max_pixsize);
            sd_wdef(qlout, INFINITE, 0, 0, &max_pixsize);
        }
#else   /* MENUS */
        /* Set window outline and size */
        iop_outl(qlout, INFINITE, 0, 0, 0, (WM_wsiz_t *) &max_pixsize);
        sd_wdef(qlout, INFINITE, 0, 0, &max_pixsize);
#endif  /* MENUS */
        mousepos = qlmouse();
#endif  /* MOUSE */
    }

    mt_dmode(&smode, &dtype);
    if (smode == 8)
        strcpy(sres, "LOW");
    else if (smode == 12)
        strcpy(sres, "THOR");
    else
        strcpy(sres, "HIGH");

#if COLOR
    qlfcol(7);                                  /* Default colours */
    qlbcol(0);
#else
    sd_setpa(qlout, INFINITE, 0);               /* Default colours */
    sd_setst(qlout, INFINITE, 0);
    sd_setin(qlout, INFINITE, 7);
#endif
    sd_clear(qlout, INFINITE);                  /* CLS */
    sd_chenq(qlout, INFINITE, &wsize);          /* Find window size in characters */
    term.t_nrow = term.t_mrow = wsize.q_height-1;       /* Define window size */
    term.t_ncol = term.t_mcol = wsize.q_width;
    ttrow = wsize.q_y;                          /* sd.chenq also tells where is */
    ttcol = wsize.q_x;                          /*  HW cursor. */

    return TRUE;
}

/*
 * This function gets called just before we go back home to the command
 * interpreter (not used under QDOS).
 */

int ttclose(void)
{
    return TRUE;
}

#if MOUSE

/* Mouse pointer routine. */

void checkmouse(void)
{
    static short    mouserow = -1;
    static short    mousecol = -1;
    static short    event    =  0;
    static struct   QLRECT size;
    int             ret;
    int             key_row;
    char            keyrow[] = { 0x09,          /* KEYROW command code.       */
                                 0x01,          /* One parameter to send.     */
                                 0x00, 0x00,    /* Send only 4 least signifi- */
                                 0x00, 0x00,    /* cant bits of the parameter.*/
                                 0x07,          /* Parameter (row 7 of kbd).  */
                                 0x02};         /* Size of reply (8 bits).    */

    if (mousepos != qlmouse() || mousebut != 0) {
        iop_rptr(qlout, 5, &mousecol, &mouserow, PT_INWIN, &mouse);
        mousepos = qlmouse();
        if (in_check())
            return;
        sd_curs(qlout, INFINITE);       /* Necessary anyway to ensure that... */
        hidden_cursor = TRUE;
#if MENUS
        if (wman && appw_pointer)
            sd_cure(qlout, INFINITE);   /* ..the cursor is actually displayed */
#endif
read_ptr:
        event = (mousebut == 0 ? (PT_KYSTK | PT_OUTWN) : PT_KEYUP);
        ret = iop_rptr(qlout, 5, &mousecol, &mouserow, event, &mouse);
        mousepos = qlmouse();
        if (in_check())
            return;
#if MENUS
        if (wman && mouse.swnr == 65535) {  /* If ptr out of app sub-window */
            if (appw_pointer) {             /* Erase the cursor as qlout is */
                sd_curs(qlout, INFINITE);   /* ...now the whole menu */
                aw0_mainmenu.pspr = NULL;   /* And back to the hand sprite */
            } else
                aw0_mainmenu.pspr = &curs_spr;
            delta_x = delta_y = 0;
read_menu:
            if (in_check())
                return;
            sd_chenq(qlout, INFINITE, &wsize);
            lit_number = -1;
            wm_rptr(&ww);
            wm_swapp(&ww, 0, 7);
            if (ws.evnt & PT_ZZZZ) {                /* Sleep event */
                killer(TRUE);
                menu_button(&ww, _prog_name);
                wm_swapp(&ww, 0, 7);
                sgarbf = TRUE;
                update(FALSE);
                mouse.kstk = 0;
            } else if (ws.evnt & PT_SCROL) {        /* Scroll event */
                wm_swapp(&ww, 0, 7);
                sd_pos(qlout, INFINITE, wsize.q_x, wsize.q_y);
                gotoline(TRUE, aw0_mainmenu.ypwcb->s[0].stt);
                update(TRUE);
                mousebut = 0;
                goto read_ptr;
            } else if (ws.evnt & PT_WSIZE) {        /* Resize event */
                sd_chenq(qlout, INFINITE, &wsize);
                mouse.kstk = 0;
            } else if (ws.evnt & PT_DO) {           /* DO event */
                if (lit_number != -1) {
                    if (ws.citem == lit_number && lit_number != 14
                        && lit_number != 15) {
                            ws.citem &= 32767;
                            wm_drbdr(&ww);
                    }
                    switch (lit_number) {
                        case 13:
                            about();
                            wm_swapp(&ww, 0, 7);
                            break;
                        case 14:
                            wm_swapp(&ww, 0, 7);
	                    sd_pos(qlout, INFINITE, wsize.q_x, wsize.q_y);
                            if (ws.kprs == 2)
                                forwpage(FALSE, 0);
                            else
                                mvdnwind(TRUE, 1);
                            update(TRUE);
                            goto read_menu;
                        case 15:
                            wm_swapp(&ww, 0, 7);
	                    sd_pos(qlout, INFINITE, wsize.q_x, wsize.q_y);
                            if (ws.kprs == 2)
                                backpage(FALSE, 0);
                            else
                                mvupwind(TRUE, 1);
                            update(TRUE);
                            goto read_menu;
                        default:
                            killer(FALSE);
                            if (getenv("EMACS_UQLXBUG"))
                                pop_menu(lit_number, NULL);
                            else
                                pop_menu(lit_number, curbp->b_fname);
                            wm_swapp(&ww, 0, 7);
                    }
                    mouse.kstk = 0;
                } else {
                    sd_pos(qlout, INFINITE, wsize.q_x, wsize.q_y);
                    if (appw_pointer)
                        sd_cure(qlout, INFINITE);
                    goto read_ptr;
                }
            }
            sd_pos(qlout, INFINITE, wsize.q_x, wsize.q_y);
            mousebut = 0;
            mousepos = qlmouse();
            return;
        } else                          /* Ptr into the application window. */
#endif  /* MENUS */
            if (ret == -1)
                goto read_ptr;

        /*
         * If the right mouse button was pressed, checks for <CTRL> key.
         * If this key is pressed, then we will pop up the context sub-menu.
         * Note that on some machines (Thor XVI) or emulators the keyrow code
         * returned could be different...
         *
         * Those with 3 buttons mouse and proper mouse driver, may pop up the
         * context sub-menu with just the third button instead of <CTRL> +
         * right mouse button.
         */
        if (wman && mouse.kprs == 2)
            key_row = mt_ipcom(keyrow);

        if (wman &&
            ((mouse.kprs == 2 && key_row == ctrl_row) || mouse.kprs == 3)) {
            /*
             * OK, so the user want a context sub-menu...
             *
             * First, put a "left mouse button down" event IOT reposition
             * the "point" once the sub-menu is called.
             */
            sd_pxenq(qlout, INFINITE, &size);
            mouserow = mouse.ypos*wsize.q_height/size.q_height;
            mousecol = mouse.xpos*wsize.q_width/size.q_width;
            event = 'a';
            in_put(0);
            in_put(MOUS >> 8);
            in_put(mousecol);
            in_put(mouserow);
            in_put(event);

            /*
             * Now we can pull down the context sub-menu, after saving
             * the cursor position and disabling it.
             */
            sd_chenq(qlout, INFINITE, &wsize);
            sd_curs(qlout, INFINITE);
            hidden_cursor = TRUE;
            killer(FALSE);
            if (getenv("EMACS_UQLXBUG"))
                pop_menu(10, NULL);
            else
                pop_menu(10, curbp->b_fname);

            /*
             * Tidy up before returning the control to MicroEMACS...
             */
            wm_swapp(&ww, 0, 7);
            sd_pos(qlout, INFINITE, wsize.q_x, wsize.q_y);
            mousebut = 0;
            mouse.kstk = 0;
        } else {
            if (mouse.kprs == 1 || mouse.kprs == 2 || mousebut != 0) {
                sd_pxenq(qlout, INFINITE, &size);
                mouserow = mouse.ypos*wsize.q_height/size.q_height;
                mousecol = mouse.xpos*wsize.q_width/size.q_width;
                event = (mouse.kprs != 0 ? 'a' : 'b');
                if (mouse.kprs == 2 || mousebut == 2)
                    event += 4;
                mousebut = mouse.kprs;
                in_put(0);
                in_put(MOUS >> 8);
                in_put(mousecol);
                in_put(mouserow);
                in_put(event);
                mouse.kstk = 0;
            } else
                mousebut = 0;

            if (mousebut == 1)
                aw0_mainmenu.pspr = &fist_spr;
            else
                if (appw_pointer)
                    aw0_mainmenu.pspr = NULL;
                else
                    aw0_mainmenu.pspr = &curs_spr;
        }
    }
    mousepos = qlmouse();
    return;
}
#endif /* MOUSE */

/*
 * Read a character from the terminal, performing no editing and doing no
 * echo at all. Also mouse events are checked here.
 *
 * EMACS likes to know about ALT keys. This doesn't stop you from using
 * ALTKEY, HOTKEY etc since all utilities of that kind 'get in' before
 * applications such as EMACS do.
 */

int qlgetc(void)
{
    unsigned char c;                        /* A character */
    char          sc;
    char          altflag = 0;              /* Was previous character an ALT */
    short         x, y;

    if (in_check())                         /* Maybe some buffered stuff */
        return in_get();

readagain:
#if MOUSE
    if (ptrenv && mouseflag) {
        while (TRUE) {
            checkmouse();
            if (in_check())
                return in_get();
            if (hidden_cursor) {
                sd_cure(qlout, INFINITE);
                hidden_cursor = FALSE;
            }
            /* Check if a keypress occured while the pointer was active.
             * If there was a keypress and it was not a wake event, then
             * take it into account, else wait for the next keypress.
             */
            if ((mouse.kstk != 0) && (mouse.kstk != 8)) {
                sc = mouse.kstk;
                mouse.kstk = 0;
                break;
            } else
                if (io_fbyte(qlout, 10, &sc) == 0)
                    break;
        }
    } else
#endif
        io_fbyte(qlout, INFINITE, &sc);     /* Fetch a character from input stream */

    c = (unsigned char) sc;

    if (altflag)                            /* MicroEMACS likes its ALT keys... */
        uppercase(&c);                      /* ...to be in UPPER CASE */

    if (c >= 0xC0 || c == 0x0A) {
        switch (c) {
            /* Some keys have to be done separately because the line input routine
             * needs to know about them.
             */
            case 0x0A:
                return 0x0D;                /* ENTER -> ^M */
                break;
            case 0xC2:
                return 0x08;                /* ^left -> ^H */
                break;
#if MENUS
            case 0xFD:                      /* SHIFT TAB -> pointer on menu */
                if (wman) {
                    x = ww.xorg + 50;
                    y = ww.yorg +  6;
                    iop_sptr(qlout, 1, &x, &y, 0);
                    mousepos++;
                    goto readagain;
                } else {
                    c = 0x09;
                    break;
                }
#endif
            case 0xFE:
                return 0x0A;                /* SHIFT ENTER  -> ^J */
                break;
            case 0xFF:                      /* Ql prefixes all ALT keys with 0xFF */
                altflag = TRUE;
                goto readagain;
            default:                        /* Convert function keys &c */
                c -= 0xC0;                  /* Get offset into table */
                if (altflag) {              /* Mask in ALT flag */
                    in_put(speclist[c] |= ALTD >> 8);
                    altflag = 0;            /* (mostly for FN keys here) */
                } else
                    in_put(speclist[c] >> 8);
                in_put(speclist[c]);
                break;
        }
        c = 0;                              /* Return an event */
    } else if (!c) {
        /* a null from kbd is event class 0, even if with ALT */
        in_put(0);
        in_put(0);

    } else if (altflag) {
        in_put(ALTD >> 8);                  /* First return ALT specification */
        in_put(c);                          /* Return the character  */
        c = 0;                              /* Return an event */
    }
    
    return c;
}

/* Screen modes. */

static struct {
    char *name;
    unsigned char value;
    } modelist[] =  {
                        "HIGH", 0,              /* Mode  4 */
                        "LOW",  8,              /* Mode  8 */
                        "THOR",12               /* Mode 12 (Thor XVI) */
                    };

#define QLMODES (sizeof modelist/sizeof modelist[0])

/* Change screen resolution. */

int qlcres(char *res)
{
    short  i;
    short  smode, dtype = -1;
    char   *p;

#if MOUSE
    if (ptrenv) {
        strcpy(sres, "HIGH");
        return TRUE;
    }
#endif

    qlflush();
    if ((p = strrchr(res, '_')) && p[1] != 0 && p[2] != 0 && p[3] == 0)
        *p++ = 0;
    for (i = 0; i < QLMODES && stricmp(res, modelist[i].name); i++)
        ;
    if (i == QLMODES)
        return FALSE;
    smode = (short) modelist[i].value;
    mt_dmode(&smode, &dtype);
    if (p)
        sd_setsz(qlout, INFINITE, p[0] - '0', p[1] - '0');
    else
        sd_setsz(qlout, INFINITE, 0, 0);
    sd_chenq(qlout, INFINITE, &wsize);  /* Find window size in characters */
    term.t_mrow = wsize.q_height-1;     /* Define maximum window size */

    term.t_mcol = wsize.q_width;
    ttrow = wsize.q_y;                  /* sd.chenq also tells where is HW cursor */
    ttcol = wsize.q_x;
    newwidth(FALSE, 0);
    newsize(FALSE, 0);
    strcpy(sres, modelist[i].name);
    return TRUE;
}

#if MENUS

static long lit_reset(struct WM_wwork *wwk, short lit)
{
    wwk->wstat->litem[lit] = WSI_MKAV;      /* set make available */
    return wm_ldraw(wwk, -1);               /* and re-draw */
}

static long AMOVE(struct WM_wwork *wwk, struct WM_litm *li)
{
    short dx, dy;

    /* Do not move anything if the screen is only 512x256 */
    if (max_pixsize.q_width == 512 && max_pixsize.q_height == 256) {
        lit_reset(wwk, li->item);
        return 0;
    }
    wwk->wstat->evnt |= PT_WMOVE;           /* set event bit for wm_chwin */
    wm_chwin(wwk, &dx, &dy);                /* move window */
    lit_reset(wwk, li->item);
    return PT__MOVE;
}

static long ASIZE(struct WM_wwork *wwk, struct WM_litm *li)
{
    /* Do not resize anything if the screen is only 512x256 */
    if (max_pixsize.q_width == 512 && max_pixsize.q_height == 256) {
        lit_reset(wwk, li->item);
        return 0;
    }

    /* If it was a do event then resize to the maximal size */
    if (ws.kprs == 2) {
        delta_x = max_pixsize.q_height;
        delta_y = -max_pixsize.q_width;
    } else {
        wwk->wstat->evnt |= PT_WSIZE;       /* set event bit for wm_chwin */
        wm_chwin(wwk, &delta_x, &delta_y);  /* resize window */
    }
    menu_resize();
    lit_reset(wwk, li->item);
    return PT__WSIZ;
}

/* Application sub-window draw action routine: draw the vertical scroll bar. */

static long ADRAW (struct WM_wwork *wwk, struct WM_appw *apw)
{
    wm_index(wwk, (struct WM_swdef *)apw);
    return 0;
}

static long ASLEEP(struct WM_wwork *wwk,  struct WM_litm *li)
{
    lit_reset(wwk, li->item);
    return PT__ZZZZ;
}

static long AHIT(struct WM_wwork *wwk, struct WM_appw *apw,
                 struct WM_wstat *wst, short x, short y, short evnt)
{
    wwk->wstat->evnt |= PT_DO;
    return 0;
}

static long lit_event(struct WM_wwork *wwk, struct WM_litm *li)
{
    lit_number = li->item;
    lit_reset(wwk, li->item);
    return PT__DO;
}

static long ACTRL(struct WM_wwork *wwk, struct WM_appw *apw,
                  struct WM_wstat *wst, short bpos, short blen,
                  short item, short evnt)
{
    int line;

    line = aw0_mainmenu.nrow * bpos / blen + 1;
    aw0_mainmenu.ypwcb->s[0].stt = line;

    ws.evnt = PT_SCROL;
    
    return PT__DO;
}

void update_scroll_bar(EWINDOW *wp)
{
    static EWINDOW *old_wp = NULL;
    LINE *lp;
    int nline = 0;
    int cline = 0;

    if (!wman || (wp != curwp))
        return;

    if ((wp != old_wp) || (wp->w_flag & ~(WFMODE | WFEDIT | WFCOLR))) {
        old_wp = wp;
        lp = lforw(wp->w_bufp->b_linep);
        do {
            if (lp == wp->w_linep)
                cline = nline;
            nline++;
        } while ((lp = lforw(lp)) != wp->w_bufp->b_linep);

        if (wp->w_linep == wp->w_bufp->b_linep)
        	cline = nline;

        aw0_mainmenu.nrow = nline;
        aw0_mainmenu.ypwcb->s[0].stt = cline;
        aw0_mainmenu.ypwcb->s[0].siz = wp->w_ntrows;

        sd_chenq(qlout, INFINITE, &wsize);
        wm_upbar(&ww, (struct WM_swdef *) &aw0_mainmenu, 0, 0);
        wm_swapp(&ww, 0, 7);
        sd_pos(qlout, INFINITE, wsize.q_x, wsize.q_y);
    }
}

static void delete_all_but_first_screen(void)
{
    SCREEN *sp;             /* ptr to current screen */

    /* Select the first screen */
    sp = first_screen;
    select_screen(sp, FALSE);

    /* Get the next screen in the list */
    sp = sp->s_next_screen;

    /* Scan and delete all the screens from there */
    while (sp) {
        unlist_screen(sp);
        free_screen(sp);
        sp = sp->s_next_screen;
    }
}

static void menu_resize(void)
{
    static WM_wsiz_t    wind;
    chanid_t            aux;
    short               xpos, ypos;

    /* First abort any command without beeping */
    if (beep_flag == TRUE)
        beep_flag = -1;
    in_put(7);
  
    /* Then select the first screen as the current screen and delete all the
       other screens. I confess that this is a rather crude method but fiddling
       around with all emacs screen, buffer and window structures IOT resize
       the whole stuff properly is just too much for me !
       At least, this method does provide a clean screen, while in the UNIX
       implementation all but the first screen is garbage !
    */
    delete_all_but_first_screen();

    /*
     * Remember the window position. IOT do this, we need an auxiliary console
     * channel.
     */

    aux = io_open("con_2x1a0x0", 0);
    iop_flim(aux, INFINITE, &wind);
    io_close(aux);
  
    /* Erase the current menu */
    wm_unset(&ww);   

    /* Reset pointer sprite in case EMACS_POINTER changed since last resize */
    if (appw_pointer)
        aw0_mainmenu.pspr = NULL;
    else
        aw0_mainmenu.pspr = &curs_spr;

    /* Adjust delta_x in case we want to enlarge beyong the screen limits */
    if (ww.xsize + delta_x > max_pixsize.q_width - 4)
        delta_x = max_pixsize.q_width - ww.xsize - 4;

    /* Make delta_x on a character width boundary */
    delta_x = 6 * (delta_x / 6);

    /* Now adjust the menu horizontal size and items placement */
    if (ww.xsize + delta_x > 508) {
        ww.xsize += delta_x;
        aw0_mainmenu.xsize += delta_x;
        infw[0].xsize += delta_x;
        infw[2].xsize += delta_x;
        litm[12].xorg += delta_x;
        litm[13].xorg += delta_x;
        litm[14].xorg += delta_x;
        litm[15].xorg += delta_x;
        infw[1].xorg   = ((ww.xsize - infw[1].xsize - 2) >> 2) * 2;
    } else {
        delta_x = ww.xsize - 508;
        ww.xsize = 508;
        aw0_mainmenu.xsize = 492;
        infw[2].xsize = 500;
        infw[0].xsize = 416;
        infw[1].xorg  = 194;
        litm[12].xorg = 486;
        litm[13].xorg = 464;
        litm[14].xorg = 472;
        litm[15].xorg = 486;
    }

    /* Adjust delta_y in case we want to enlarge beyong the screen limits */
    if (ww.ysize - delta_y > max_pixsize.q_height - 2)
        delta_y = ww.ysize - max_pixsize.q_height + 2;

    /* Make delta_y on a character height boundary */
    delta_y = 10 * (delta_y / 10);

    /* Now adjust the menu vertical size */
    if (ww.ysize - delta_y > 254) {
        ww.ysize -= delta_y;
        aw0_mainmenu.ysize -= delta_y;
    } else {
        delta_y = ww.ysize - 254;
        ww.ysize = 254;
        aw0_mainmenu.ysize = 222;
    }

    /*
     * Compute the new upper left corner position (taking the border width
     * into account).
     */

    xpos = wind.xorg + 2;
    ypos = wind.yorg + delta_y + 1;
    if (ypos <= 0)
        ypos = 1;

    ww.xorg = ww.yorg = 0;  /* Needed as the menu pos is relative to ptr pos */

    /* Redraw the resized menu */
    wm_prpos(&ww, xpos, ypos);
    wm_wdraw(&ww);

    /* Find new application sub-window size in characters */
    wm_swapp(&ww, 0, 7);
    sd_chenq(qlout, INFINITE, &wsize);

    /* Persuade MicroEMACS to redraw all its stuff properly */
    winch_vtresize(wsize.q_height, wsize.q_width);
    onlywind(FALSE, 0);
    sgarbf = TRUE;
    update(FALSE);
}

static void about(void)
{
    /* Window status area */

    static struct WM_wstat wsa = {
        NULL,               /* *wwork               point back to work def */
        NULL,               /* *wdef                this was wdef */
        0,                  /* chid                 chan ID enclosing pointer */
        0,0,0,              /* swnr, xpos, ypos     sub-window and ptr posn */
        0,0,                /* kstk, kprs           key stroke and press */
        0,                  /* evnt                 event vector */
        0,0,0,0,            /* xsiz, ysiz, xorg, yorg  */
        0,0,                /* ptpsx, ptpsy         pointer position (abs) */
        0,0,0,              /* wmode, 2 spare       window mode and spare */
        NULL,               /* *ciact               current item "action" */
        -1,                 /* citem                current item */
        0,0,0,              /* cibrw, cipap, cispr  current item attr */
        0,0,0,0,            /* cihxs,ys, cihxo,yo   current item hit area */
        {0,0,0,0,0,0,0,0}   /* litem[40]            loose item status */
    };

    #define ninfwa   4    /* number of information windows */
    #define ninfoa   9    /* number of information objects */
    #define nlitma   1    /* number of loose items */
    #define nappla   0    /* number of application sub-windows */

    /* Text objects */

    static QD_TEXTI(ob1_about, PROGNAME" v"VERSION"."PATCHLEVEL" ("COMPILER")");
    static QD_TEXTI(ob2_about, "for QDOS/SMS systems");
    static QD_TEXTI(ob3_about, "written by Daniel M. Lawrence");
    static QD_TEXTI(ob4_about, "[based on code by Dave Conroy]");
    static QD_TEXTI(ob5_about, "QDOS/SMS port (" __DATE__ ") by:");
    static QD_TEXTI(ob6_about, "- Richard Kettlewell 1991-1992 (v3.11)");
    static QD_TEXTI(ob7_about, "- Thierry Godefroy   1993-2000 (v3.11");
    static QD_TEXTI(ob8_about, "  & v4.00 patch levels 00 to " PATCHLEVEL ").");
    static QD_TEXTI(ob9_about, "About");
    static QD_TEXTI(ob10_about, "OK");

    /* Information objects list for information window 0 */

    static struct WM_info ol0_about[8+1] = {
        {180, 12,                   /* xsize, ysize      size */
          28,  3,                   /* xorg, yorg        and origin */
           0,  0,                   /* type, spare       item type */
         {7,0,0},                   /* ink, cwid, chgt   character attributes */
         &ob1_about},               /* *pobj             pointer to object */

        {126, 12,                   /* xsize, ysize      size */
          84, 15,                   /* xorg, yorg        and origin */
           0,  0,                   /* type, spare       item type */
         {7,0,0},                   /* ink, cwid, chgt   character attributes */
         &ob2_about},               /* *pobj             pointer to object */

        {180, 12,                   /* xsize, ysize      size */
          30, 33,                   /* xorg, yorg        and origin */
           0,  0,                   /* type, spare       item type */
         {4,0,0},                   /* ink, cwid, chgt   character attributes */
         &ob3_about},               /* *pobj             pointer to object */

        {180, 12,                   /* xsize, ysize      size */
          28, 45,                   /* xorg, yorg        and origin */
           0,  0,                   /* type, spare       item type */
         {4,0,0},                   /* ink, cwid, chgt   character attributes */
         &ob4_about},               /* *pobj             pointer to object */

        {232, 12,                   /* xsize, ysize      size */
           4, 63,                   /* xorg, yorg        and origin */
           0,  0,                   /* type, spare       item type */
         {7,0,0},                   /* ink, cwid, chgt   character attributes */
         &ob5_about},               /* *pobj             pointer to object */

        {232, 12,                   /* xsize, ysize      size */
           4, 81,                   /* xorg, yorg        and origin */
           0,  0,                   /* type, spare       item type */
         {4,0,0},                   /* ink, cwid, chgt   character attributes */
         &ob6_about},               /* *pobj             pointer to object */

        {232, 12,                   /* xsize, ysize      size */
           4, 93,                   /* xorg, yorg        and origin */
           0,  0,                   /* type, spare       item type */
         {4,0,0},                   /* ink, cwid, chgt   character attributes */
         &ob7_about},               /* *pobj             pointer to object */

        {232, 12,                   /* xsize, ysize      size */
           6,105,                   /* xorg, yorg        and origin */
           0,  0,                   /* type, spare       item type */
         {4,0,0},                   /* ink, cwid, chgt   character attributes */
         &ob8_about},               /* *pobj             pointer to object */

        {-1}
    };


    /* Information objects list for information window 2 */

    static struct WM_info ol2_about[1+1] = {
        { 34, 10,                   /* xsize, ysize      size */
           4,  2,                   /* xorg, yorg        and origin */
           0,  0,                   /* type, spare       item type */
         {7,0,0},                   /* ink, cwid, chgt   character attributes */
         &ob9_about},               /* *pobj             pointer to object */

        {-1}
    };

    /* Information windows list */

   static struct WM_infw infwa[ninfwa+1] = {
        {240,133,                   /* xsize, ysize  */
           4, 14,                   /* xorg, yorg */
         0,                         /* flag */
         1,2,0,                     /* Borw, Borc,papc    character attributes */
         ol0_about},                /* *pobj             pointer to object */

        {248, 11,                   /* xsize, ysize  */
           0,  1,                   /* xorg, yorg */
         0,                         /* flag */
         0,0,100,                   /* Borw, Borc,papc    character attributes */
         NULL},                     /* *pobj             pointer to object */

        { 38, 12,                   /* xsize, ysize  */
         104,  0,                   /* xorg, yorg */
         0,                         /* flag */
         0,0,0,                     /* Borw, Borc,papc    character attributes */
         ol2_about},                /* *pobj             pointer to object */

        {240,  2,                   /* xsize, ysize  */
           4,131,                   /* xorg, yorg */
         0,                         /* flag */
         0,0,80,                    /* Borw, Borc,papc    character attributes */
         NULL},                     /* *pobj             pointer to object */

        {-1}
    };

    /* Loose items list */

    static struct WM_litm litma[nlitma+1] = {
        {236, 12,                   /* xsize, ysize  */
           6,134,                   /* xorg, yorg */
           0,  0,                   /* xjst,yjst */
          -1,'O',                   /* Type ,Skey */
         &ob10_about,               /* Pointer to object */
           0,                       /* Item Number */
         &afun},                    /* *pact   pointer to action */

        {-1}
    };

    /* Window working definition */

    static struct WM_wwork wwa = {
        &wsa,               /* *wstat;              pointer to status area */
        NULL,               /* *wscale              this was pointer to WDEF */
        0,                  /* chid                 channel ID */
        NULL,               /* *pprec               pointer to pointer record */
        0,                  /* psave                pointer position saved */
        0,0,0,0,            /* 4 zero */
        NULL,               /* *splst               pointer to sprite list */
        248,149,            /* xsize, ysize         size */
        0, 0,               /* xorg, yorg           and origin */
        2,                  /* flag */
        1,7,0,              /* borw, borc, papr     border width/clr, paper */
        &hand_spr,          /* *sprite              pointer to sprite */
        1,4,                /* curw, curc           current item width/clr */
        0,2,NULL,NULL,      /* uback, uink, *ublob, *upatt  unavail */
        0,7,NULL,NULL,      /* aback, aink, *ablob, *apatt  avail */
        2,7,NULL,NULL,      /* sback, sink, *sblob, *spatt  selected */
        NULL,               /* *help */
        ninfwa,ninfoa,infwa,    /* ninfo, ninob, *pinfo info winds */
        nlitma,litma,           /* nlitm, *plitm        loose items */
        nappla,NULL             /* nappl  *pappl        appl winds */
    };

    wwa.chid = io_open("con_2x1a0x0", 0);

    if(!wm_findv(wwa.chid))
        exit (-1);

    wsa.wwork=&wwa;
    wm_pulld(&wwa, -1, -1);
    wm_wdraw(&wwa);
    wm_rptr(&wwa);
    wm_unset(&wwa);
    io_close(wwa.chid);
}

#endif  /* MENUS */

/* Change palette - meaningless on Ql. */

int spal(char *pstr)
{
return TRUE;
}

/* Make a noise. */

int qlbeep(void)
{
#if NOISY
    qlflush();
    if (beep_flag == TRUE)
        beep(beep_dur, beep_pitch);
    if (beep_flag == -1)
        beep_flag = TRUE;
#endif
}

/* Pause the program for pause/50 seconds */

int qlsleep(int pause)
{
    mt_susjb(-1, pause, NULL);
}

/* Label a function key - meaningless on Ql. */

#if FLABEL
int fnclabel(int f, int n)
{
    return TRUE;
}
#endif

#else
qlhello() { }
#endif /* QDOS */
